#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

def write2bxsf ( data_controller, fname, bands, nbnd, indices, fermi_up, fermi_dw ):
  '''
  Writes a bxsf File to 'opath' (outputdir)

  Arguments:
      data_controller (DataController): The current PAOFLOW DataController
      fname (str): File name
      bands (ndarray): Array with data to write
      nbnd (int): Number of columns of band-like data values to write
      indices (list): A list of included band indices, in ascending order
      fermi_up (float): Highest acceptable energy
      fermi_dw (float): Lowest acceptable energy

  Returns:
      None
  '''
  import numpy as np
  from os.path import join

  pi = np.pi
  arrays,attributes = data_controller.data_dicts()

  x0 = np.zeros(3, dtype=float)
  if indices is None:
    indices = np.zeros(nbnd, dtype=float)

  Efermi = attributes['Efermi']
  alat,b_vectors = attributes['alat'],arrays['b_vectors']
  nx,ny,nz = attributes['nk1'],attributes['nk2'],attributes['nk3']

  with open (join(attributes['workpath'],attributes['outputdir'],'{0}'.format(fname)),'w') as f:
    f.write('\nBEGIN_INFO\n  Fermi Energy: {:15.9f}\n  Shift Range: {:12.9f}eV to{:12.9f}eV\nEND_INFO\n'.format(Efermi,fermi_dw,fermi_up))
    # BXSF scalar-field header
    f.write('\nBEGIN_BLOCK_BANDGRID_3D\nband_energies\nBANDGRID_3D_BANDS\n')  
    # number of points in each direction
    f.write('{:12d}\n'.format(nbnd))
    f.write('{:12d}{:12d}{:12d}\n'.format(nx+1,ny+1,nz+1))
    # origin (should be zero, if I understan correctly)
    f.write('  {}\n'.format(''.join('%10.6f'%F for F in x0)))
    # 1st spanning (=lattice) vector
    f.write('  {}\n'.format(''.join('%10.6f'%F for F in b_vectors[0]*2*pi/alat)))
    # 2nd spanning (=lattice) vector
    f.write('  {}\n'.format(''.join('%10.6f'%F for F in b_vectors[1]*2*pi/alat)))
    # 3rd spanning (=lattice) vector
    f.write('  {}\n'.format(''.join('%10.6f'%F for F in b_vectors[2]*2*pi/alat)))

    for ib in range(nbnd):
      f.write('  BAND: {:5d}\n'.format(int(indices[ib])+1))
      for ix in range(nx):
        for iy in range(ny):
          f.write('    {}'.format(''.join('%15.9f'%F for F in bands[ix,iy,:,ib])))
          f.write('{:15.9f}\n'.format(bands[ix,iy,0,ib]))
        f.write('    {}'.format(''.join('%15.9f'%F for F in bands[ix,0,:,ib])))
        f.write('{:15.9f}\n'.format(bands[ix,0,0,ib]))
      for iy in range(ny):
        f.write('    {}'.format(''.join('%15.9f'%F for F in bands[0,iy,:,ib])))
        f.write('{:15.9f}\n'.format(bands[0,iy,0,ib]))
      f.write('    {}'.format(''.join('%15.9f'%F for F in bands[0,0,:,ib])))
      f.write('{:15.9f}\n'.format(bands[0,0,0,ib]))        
    f.write('END_BANDGRID_3D\nEND_BLOCK_BANDGRID_3d\n')



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from scipy import linalg as spl
from numpy import linalg as npl

def bands_calc ( data_controller ):
  from .communication import scatter_full, gather_full

  arrays,attributes = data_controller.data_dicts()

  npool = attributes['npool']
  nawf,_,nk1,nk2,nk3,nspin = arrays['HRs'].shape

  kq_aux = scatter_full(arrays['kq'].T, npool).T
 
  Hks_aux = band_loop_H(data_controller, kq_aux)

  E_kp_aux = np.zeros((kq_aux.shape[1],nawf,nspin), dtype=float, order="C")
  v_kp_aux = np.zeros((kq_aux.shape[1],nawf,nawf,nspin), dtype=complex, order="C")

  for ispin in range(nspin):
    for ik in range(kq_aux.shape[1]):
      E_kp_aux[ik,:,ispin],v_kp_aux[ik,:,:,ispin] = spl.eigh(Hks_aux[:,:,ik,ispin], b=(None), lower=False, overwrite_a=True, overwrite_b=True, turbo=True, check_finite=True)

  Hks_aux = Sks_aux = None
  return E_kp_aux, v_kp_aux


def band_loop_H ( data_controller, kq_aux ):

  arrays,attributes = data_controller.data_dicts()

  nksize = kq_aux.shape[1]
  nawf,_,nk1,nk2,nk3,nspin = arrays['HRs'].shape

  HRs = np.reshape(arrays['HRs'], (nawf,nawf,nk1*nk2*nk3,nspin), order='C')
  kdot = np.tensordot(arrays['R'], 2.0j*np.pi*kq_aux, axes=([1],[0]))
  np.exp(kdot, kdot)
  Haux = np.zeros((nawf,nawf,nksize,nspin), dtype=complex, order="C")

  for ispin in range(nspin):
    Haux[:,:,:,ispin] = np.tensordot(HRs[:,:,:,ispin], kdot, axes=([2],[0]))

  kdot  = None
  return Haux


def do_bands ( data_controller ):
  from mpi4py import MPI
  from .constants import ANGSTROM_AU
  from .communication import gather_full
  from .get_R_grid_fft import get_R_grid_fft
  from .kpnts_interpolation_mesh import kpnts_interpolation_mesh

  rank = MPI.COMM_WORLD.Get_rank()

  arrays,attributes = data_controller.data_dicts()

  # Bohr to Angstrom
  attributes['alat'] /= ANGSTROM_AU

#### What about ONEDIM?
  attributes['onedim'] = False
  if not attributes['onedim']:
    #--------------------------------------------
    # Compute bands on a selected path in the BZ
    #--------------------------------------------

    alat = attributes['alat']
    nawf,_,nk1,nk2,nk3,nspin = arrays['HRs'].shape
    nktot = nk1*nk2*nk3

    # Define real space lattice vectors
    get_R_grid_fft(data_controller, nk1, nk2, nk3)

    # Define k-point mesh for bands interpolation
    kpnts_interpolation_mesh(data_controller)

    nkpi = arrays['kq'].shape[1]
    for n in range(nkpi):
      arrays['kq'][:,n] = np.dot(arrays['kq'][:,n], arrays['b_vectors'])

    # Compute the bands along the path in the IBZ
    arrays['E_k'],arrays['v_k'] = bands_calc(data_controller)

#### 1D Bands not implemented
  else:
    if rank == 0:
      print('OneDim bands not implemented in PAOFLOW_Class')
      #----------------------
      # FFT interpolation along a single directions in the BZ
      #----------------------
      #if rank == 0 and verbose: print('... computing bands along a line')
      #if rank == 0: bands_calc_1D(Hks,inputpath)

  # Angstrom to Bohr
  attributes['alat'] *= ANGSTROM_AU



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import cmath
import numpy as np

def do_spin_orbit_H ( data_controller ):

  # construct TB spin orbit Hamiltonian (following Abate and Asdente, Phys. Rev. 140, A1303 (1965))

  arry,attr = data_controller.data_dicts()

  natoms = attr['natoms']
  theta,phi = attr['theta'],attr['phi']
  norb_array,orb_pseudo = arry['naw'],arry['orb_pseudo']

  nawf,_,nk1,nk2,nk3,nspin = arry['HRs'].shape

  socStrengh = np.zeros((natoms,2), dtype=float)
  socStrengh[:,0] = arry['lambda_p'][:]
  socStrengh[:,1] = arry['lambda_d'][:]

  HR_double = np.zeros((2*nawf,2*nawf,nk1,nk2,nk3,nspin), dtype=complex)

  # nonmagnetic :  copy H at the upper (lower) left (right) of the double matrix HR_double
  if nspin == 1:
    HR_double[:nawf,:nawf,:,:,:,0] = HR_double[nawf:2*nawf,nawf:2*nawf,:,:,:,0] = arry['HRs'][:nawf,:nawf,:,:,:,0]
  # magnetic :  copy H_up (H_down) at the upper (lower) left (right) of the double matrix 
  else:
    HR_double[:nawf,:nawf,:,:,:,0] = arry['HRs'][:nawf,:nawf,:,:,:,0]
    HR_double[nawf:2*nawf,nawf:2*nawf,:,:,:,0] = arry['HRs'][:nawf,:nawf,:,:,:,1]

  ##### Need clarification
  offset = np.zeros(natoms, dtype=int)
  for i in range(1, natoms):
    offset[i] = offset[i-1] + norb_array[i-1]

  for n,norb in enumerate(norb_array):
    HR_soc_p = np.zeros((2*norb,2*norb), dtype=complex)  
    HR_soc_d = np.zeros((2*norb,2*norb), dtype=complex)

    if (orb_pseudo[n] == 's'):
      pass
    if (orb_pseudo[n] == 'sp'):
      HR_soc_p[:,:] = soc_p_sp(theta, phi, norb)
    if (orb_pseudo[n] == 'spd'):
      HR_soc_p[:,:] = soc_p_spd(theta, phi, norb)
      HR_soc_d[:,:] = soc_d_spd(theta, phi, norb)
    if (orb_pseudo[n] == 'ps'):
      HR_soc_p[:,:] = soc_p_ps(theta, phi, norb)
    if (orb_pseudo[n] == 'sspd'):
      HR_soc_p[:,:] = soc_p_sspd(theta, phi, norb)
      HR_soc_d[:,:] = soc_d_sspd(theta, phi, norb)
    if (orb_pseudo[n] == 'ssp'):
      pass
    if (orb_pseudo[n] == 'ssppd'):
      HR_soc_p[:,:] = soc_p_ssppd(theta, phi, norb)
      HR_soc_d[:,:] = soc_d_ssppd(theta, phi, norb)

    uui = offset[n]
    uuj = uui + norb
    duj = udi = ddi = uui + nawf
    dui = udj = ddj = uuj + nawf

    # Up-Up
    HR_double[uui:uuj,uui:uuj,0,0,0,0] += socStrengh[n,0]*HR_soc_p[0:norb,0:norb] + socStrengh[n,1]*HR_soc_d[0:norb,0:norb]
    # Down-Down
    HR_double[ddi:ddj,ddi:ddj,0,0,0,0] += socStrengh[n,0]*HR_soc_p[norb:2*norb,norb:2*norb]  + socStrengh[n,1]*HR_soc_d[norb:2*norb,norb:2*norb]
    # Up-Down
    HR_double[uui:uuj,udi:udj,0,0,0,0] += socStrengh[n,0]*HR_soc_p[0:norb,norb:2*norb] + socStrengh[n,1]*HR_soc_d[0:norb,norb:2*norb]
    # Down-Up
    HR_double[ddi:ddj,uui:uuj,0,0,0,0] += socStrengh[n,0]*HR_soc_p[norb:2*norb,0:norb] + socStrengh[n,1]*HR_soc_d[norb:2*norb,0:norb]

    del arry['HRs']
    arry['HRs'] = HR_double
    attr['nawf'] = arry['HRs'].shape[0]


################### PSEUDOPOTENTIAL PS ##############################33
def soc_p_ps ( theta, phi, norb ):

  HR_soc = np.zeros((2*norb,2*norb), dtype=complex) 

  sTheta,sPhi = np.sin(theta),np.sin(phi)
  cTheta,cPhi = np.cos(theta),np.cos(phi)

  #Spin Up - Spin Up  part of the p-satets Hamiltonian
  HR_soc[0,1] = -0.5j * sTheta * sPhi
  HR_soc[0,2] =  0.5j * sTheta * cPhi
  HR_soc[1,2] = -0.5j * cTheta
  HR_soc[1,0] = np.conj(HR_soc[0,1])
  HR_soc[2,0] = np.conj(HR_soc[0,2])
  HR_soc[2,1] = np.conj(HR_soc[1,2])
  #Spin Down - Spin Down  part of the p-satets Hamiltonian
  HR_soc[5:8,5:8] = -HR_soc[1:4,1:4] 
  #Spin Up - Spin Down  part of the p-satets Hamiltonian
  HR_soc[0,5] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[0,6] = -0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[1,6] =  0.5j * sTheta
  HR_soc[1,4] = -HR_soc[0,5]
  HR_soc[2,4] = -HR_soc[0,6]
  HR_soc[2,5] = -HR_soc[1,6]
  #Spin Down - Spin Up  part of the p-satets Hamiltonian
  HR_soc[5,0] = np.conj(HR_soc[0,5])
  HR_soc[6,0] = np.conj(HR_soc[0,6])
  HR_soc[4,1] = np.conj(HR_soc[1,4])
  HR_soc[6,1] = np.conj(HR_soc[1,6])
  HR_soc[4,2] = np.conj(HR_soc[2,4])
  HR_soc[5,2] = np.conj(HR_soc[2,5])

  return HR_soc
################## END PSEUDOPOTENTIAL PS ##############################


################## PSEUDOPOTENTIAL SP ##############################
def soc_p_sp ( theta, phi, norb):

  HR_soc = np.zeros((2*norb,2*norb), dtype=complex) 

  sTheta,sPhi = np.sin(theta),np.sin(phi)
  cTheta,cPhi = np.cos(theta),np.cos(phi)

  # Spin Up - Spin Up  part of the p-satets Hamiltonian
  HR_soc[1,2] = -0.5j * sTheta * sPhi
  HR_soc[1,3] =  0.5j * sTheta * cPhi
  HR_soc[2,3] = -0.5j * cTheta
  HR_soc[2,1] = np.conj(HR_soc[1,2])
  HR_soc[3,1] = np.conj(HR_soc[1,3])
  HR_soc[3,2] = np.conj(HR_soc[2,3])
  # Spin Down - Spin Down  part of the p-satets Hamiltonian
  HR_soc[5:8,5:8] = -HR_soc[1:4,1:4] 
  # Spin Up - Spin Down  part of the p-satets Hamiltonian
  HR_soc[1,6] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[1,7] = -0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[2,7] =  0.5j * sTheta
  HR_soc[2,5] = -HR_soc[1,6]
  HR_soc[3,5] = -HR_soc[1,7]
  HR_soc[3,6] = -HR_soc[2,7]
  # Spin Down - Spin Up  part of the p-satets Hamiltonian
  HR_soc[6,1] = np.conj(HR_soc[1,6])
  HR_soc[7,1] = np.conj(HR_soc[1,7])
  HR_soc[5,2] = np.conj(HR_soc[2,5])
  HR_soc[7,2] = np.conj(HR_soc[2,7])
  HR_soc[5,3] = np.conj(HR_soc[3,5])
  HR_soc[6,3] = np.conj(HR_soc[3,6])

  return HR_soc
################## END PSEUDOPOTENTIAL SP ##############################


################## PSEUDOPOTENTIAL SPD ##############################
def soc_p_spd ( theta, phi, norb ):

  HR_soc = np.zeros((2*norb,2*norb), dtype=complex) 

  sTheta,sPhi = np.sin(theta),np.sin(phi)
  cTheta,cPhi = np.cos(theta),np.cos(phi)

  # Spin Up - Spin Up  part of the p-satets Hamiltonian
  HR_soc[1,2] = -0.5j * sTheta * sPhi
  HR_soc[1,3] =  0.5j * sTheta * cPhi
  HR_soc[2,3] = -0.5j * cTheta
  HR_soc[2,1] = np.conj(HR_soc[1,2])
  HR_soc[3,1] = np.conj(HR_soc[1,3])
  HR_soc[3,2] = np.conj(HR_soc[2,3])
  #Spin Down - Spin Down  part of the p-satets Hamiltonian
  HR_soc[10:13,10:13] = -HR_soc[1:4,1:4] 
  # Spin Up - Spin Down  part of the p-satets Hamiltonian
  HR_soc[1,11] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[1,12] = -0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[2,12] =  0.5j * sTheta
  HR_soc[2,10] = -HR_soc[1,11]
  HR_soc[3,10] = -HR_soc[1,12]
  HR_soc[3,11] = -HR_soc[2,12]
  # Spin Down - Spin Up  part of the p-satets Hamiltonian
  HR_soc[11,1] = np.conj(HR_soc[1,11])
  HR_soc[12,1] = np.conj(HR_soc[1,12])
  HR_soc[10,2] = np.conj(HR_soc[2,10])
  HR_soc[12,2] = np.conj(HR_soc[2,12])
  HR_soc[10,3] = np.conj(HR_soc[3,10])
  HR_soc[11,3] = np.conj(HR_soc[3,11])

  return HR_soc


def soc_d_spd(theta,phi,norb):

  HR_soc = np.zeros((2*norb,2*norb), dtype=complex) 

  sTheta,sPhi = np.sin(theta),np.sin(phi)
  cTheta,cPhi = np.cos(theta),np.cos(phi)

  s3 = np.sqrt(3.0)

  # Spin Up - Spin Up  part of the d-satets Hamiltonian
  HR_soc[4,5] = -s3 * 0.5j * sTheta * sPhi
  HR_soc[4,6] =  s3 * 0.5j * sTheta * cPhi
  HR_soc[5,6] = -0.5j * cTheta
  HR_soc[5,7] = -0.5j * sTheta * sPhi
  HR_soc[5,8] =  0.5j * sTheta * cPhi
  HR_soc[6,7] = -0.5j * sTheta * cPhi
  HR_soc[6,8] = -0.5j * sTheta * sPhi
  HR_soc[7,8] = -1.0j * cTheta
  HR_soc[5,4] = np.conj(HR_soc[4,5])
  HR_soc[6,4] = np.conj(HR_soc[4,6])
  HR_soc[6,5] = np.conj(HR_soc[5,6])
  HR_soc[7,5] = np.conj(HR_soc[5,7])
  HR_soc[8,5] = np.conj(HR_soc[5,8])
  HR_soc[7,6] = np.conj(HR_soc[6,7])
  HR_soc[8,6] = np.conj(HR_soc[6,8])

  # Spin Down - Spin Down  part of the p-satets Hamiltonian
  HR_soc[13:18,13:18] = -HR_soc[4:9,4:9] 
  # Spin Up - Spin Down  part of the p-satets Hamiltonian
  HR_soc[4,14] = -s3 * 0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[4,15] = -s3 * 0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[5,15] =  0.5j * sTheta
  HR_soc[5,16] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[5,17] = -0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[6,16] =  0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[6,17] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[7,17] =  1j * sTheta
  HR_soc[5,13] = -HR_soc[4,14] 
  HR_soc[6,13] = -HR_soc[4,15] 
  HR_soc[6,14] = -HR_soc[5,15]
  HR_soc[7,14] = -HR_soc[5,16] 
  HR_soc[8,14] = -HR_soc[5,17] 
  HR_soc[7,15] = -HR_soc[6,16] 
  HR_soc[8,15] = -HR_soc[6,17] 
  HR_soc[8,16] = -HR_soc[7,17] 
  # Spin Down - Spin Up  part of the p-satets Hamiltonian
  HR_soc[14,4] = np.conj(HR_soc[4,14]) 
  HR_soc[15,4] = np.conj(HR_soc[4,15])
  HR_soc[15,5] = np.conj(HR_soc[5,15])   
  HR_soc[16,5] = np.conj(HR_soc[5,16])   
  HR_soc[17,5] = np.conj(HR_soc[5,17])   
  HR_soc[16,6] = np.conj(HR_soc[6,16]) 
  HR_soc[17,6] = np.conj(HR_soc[6,17])    
  HR_soc[17,7] = np.conj(HR_soc[7,17])    
  HR_soc[13,5] = np.conj(HR_soc[5,13])
  HR_soc[13,6] = np.conj(HR_soc[6,13])
  HR_soc[14,6] = np.conj(HR_soc[6,14])
  HR_soc[14,7] = np.conj(HR_soc[7,14])
  HR_soc[14,8] = np.conj(HR_soc[8,14])
  HR_soc[15,7] = np.conj(HR_soc[7,15])
  HR_soc[15,8] = np.conj(HR_soc[8,15])

  return HR_soc
################## END PSEUDOPOTENTIAL SPD ##############################


################## PSEUDOPOTENTIAL SSPD ##############################
def soc_p_sspd ( theta, phi, norb ):

  # Hardcoded to s,p,d. This must change latter.
  HR_soc = np.zeros((2*norb,2*norb), dtype=complex) 

  sTheta,sPhi = np.sin(theta),np.sin(phi)
  cTheta,cPhi = np.cos(theta),np.cos(phi)

  # Spin Up - Spin Up  part of the p-satets Hamiltonian
  HR_soc[2,3] = -0.5j * sTheta*sPhi
  HR_soc[2,4] =  0.5j * sTheta*cPhi
  HR_soc[3,4] = -0.5j * cTheta
  HR_soc[3,2] = np.conj(HR_soc[2,3])
  HR_soc[4,2] = np.conj(HR_soc[2,4])
  HR_soc[4,3] = np.conj(HR_soc[3,4])
  # Spin Down - Spin Down  part of the p-satets Hamiltonian
  HR_soc[12:15,12:15] = -HR_soc[2:5,2:5] 
  # Spin Up - Spin Down  part of the p-satets Hamiltonian
  HR_soc[2,13] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[2,14] = -0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[3,14] =  0.5j * sTheta
  HR_soc[3,12] = -HR_soc[2,13]
  HR_soc[4,12] = -HR_soc[2,14]
  HR_soc[4,13] = -HR_soc[3,14]
  # Spin Down - Spin Up  part of the p-satets Hamiltonian
  HR_soc[13,2] = np.conj(HR_soc[2,13])
  HR_soc[14,2] = np.conj(HR_soc[2,14])
  HR_soc[12,3] = np.conj(HR_soc[3,12])
  HR_soc[14,3] = np.conj(HR_soc[3,14])
  HR_soc[12,4] = np.conj(HR_soc[4,12])
  HR_soc[13,4] = np.conj(HR_soc[4,13])

  return HR_soc


def soc_d_sspd ( theta, phi, norb ):

  # Hardcoded to s,p,d. This must change latter.
  HR_soc = np.zeros((2*norb,2*norb), dtype=complex)

  sTheta,sPhi = np.sin(theta),np.sin(phi)
  cTheta,cPhi = np.cos(theta),np.cos(phi)

  s3 = cmath.sqrt(3.0)

  #Spin Up - Spin Up  part of the d-satets Hamiltonian
  HR_soc[5,6] = -s3 * 0.5j * sTheta * sPhi
  HR_soc[5,7] =  s3 * 0.5j * sTheta * cPhi
  HR_soc[6,7] = -0.5j * cTheta
  HR_soc[6,8] = -0.5j * sTheta * sPhi
  HR_soc[6,9] =  0.5j * sTheta * cPhi
  HR_soc[7,8] = -0.5j * sTheta * cPhi
  HR_soc[7,9] = -0.5j * sTheta * sPhi
  HR_soc[8,9] = -1j * cTheta
  HR_soc[6,5] = np.conj(HR_soc[5,6])
  HR_soc[7,5] = np.conj(HR_soc[5,7])
  HR_soc[7,6] = np.conj(HR_soc[6,7])
  HR_soc[8,6] = np.conj(HR_soc[6,8])
  HR_soc[9,6] = np.conj(HR_soc[6,9])
  HR_soc[8,7] = np.conj(HR_soc[7,8])
  HR_soc[9,7] = np.conj(HR_soc[7,9])

  # Spin Down - Spin Down  part of the p-satets Hamiltonian
  HR_soc[15:20,15:20] = -HR_soc[5:10,5:10] 
  # Spin Up - Spin Down  part of the p-satets Hamiltonian
  HR_soc[5,16] = -s3 * 0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[5,17] = -s3 * 0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[6,17] =  0.5j * sTheta
  HR_soc[6,18] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[6,19] = -0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[7,18] =  0.5 * complex(sPhi, -cTheta*cPhi)
  HR_soc[7,19] = -0.5 * complex(cPhi, cTheta*sPhi)
  HR_soc[8,19] =  1j * sTheta
  HR_soc[6,15] = -HR_soc[5,16] 
  HR_soc[7,15] = -HR_soc[5,17] 
  HR_soc[7,16] = -HR_soc[6,17]
  HR_soc[8,16] = -HR_soc[6,18] 
  HR_soc[9,16] = -HR_soc[6,19] 
  HR_soc[8,17] = -HR_soc[7,18] 
  HR_soc[9,17] = -HR_soc[7,19] 
  HR_soc[9,18] = -HR_soc[8,19] 
  # Spin Down - Spin Up  part of the p-satets Hamiltonian
  HR_soc[16,5] = np.conj(HR_soc[5,16]) 
  HR_soc[17,5] = np.conj(HR_soc[5,17])
  HR_soc[17,6] = np.conj(HR_soc[6,17])   
  HR_soc[18,6] = np.conj(HR_soc[6,18])   
  HR_soc[19,6] = np.conj(HR_soc[6,19])   
  HR_soc[18,7] = np.conj(HR_soc[7,18]) 
  HR_soc[19,7] = np.conj(HR_soc[7,19])    
  HR_soc[19,8] = np.conj(HR_soc[8,19])    
  HR_soc[15,6] = np.conj(HR_soc[6,15])
  HR_soc[15,7] = np.conj(HR_soc[7,15])
  HR_soc[16,7] = np.conj(HR_soc[7,16])
  HR_soc[16,8] = np.conj(HR_soc[8,16])
  HR_soc[16,9] = np.conj(HR_soc[9,16])
  HR_soc[17,8] = np.conj(HR_soc[8,17])
  HR_soc[17,9] = np.conj(HR_soc[9,17])

  return HR_soc
 ################## END PSEUDOPOTENTIAL SSPD ##############################


#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016,2017 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

# this version works only for non-magnetic or non-collienar calculations
def wave_function_site_projection(data_controller):
    import cmath
    import numpy as np
    from mpi4py import MPI
    from os.path import join
    from scipy import fftpack as FFT
    from constants import ANGSTROM_AU
    from scipy.fftpack import fftshift
    from communication import scatter_full,gather_full
    
    arry,attr = data_controller.data_dicts()

    tau = arry['tau'] / ANGSTROM_AU
    naw,v_k = arry['naw'],array['v_k']   
    bands,k_index = arry['bands_proj'],attr['k_proj']   
    
    # sites to project the wave function
    site = np.copy(tau[:,0])

    do_spin_orbit = attr['do_spin_orbit']    
    nawf,dim = attr['nawf'],attr['dimension']

    for idx in range(bands.shape[0]):
        bnd_idx=bands[idx]  # index of the band to be projected
        # open file
        f=open(join(attr['opath'],'site-projected-wave-function-'+str(bnd_idx)+'.dat'), 'w')
        for n in range(tau.shape[0]):

            # creating masks to consirer only the n site.
            mask   = np.zeros_like(v_k[k_index:k_index+1,:,:,0])
            cs     = np.zeros_like(v_k[k_index:k_index+1,:,:,0])

            # seting up the nonzero parts of the mask and the wave-function
            idx= np.sum(naw[0:n]) # initial
            fdx= idx + naw[n]     # final

            # Do to the doubling of the Hamiltonian when SOC is included in the PAO Hamiltonian.
            if (do_spin_orbit):
                s=int(nawf/2)
                mask[:,idx:fdx,:]    =  complex(1.0, 1.0)
                mask[:,idx+s:fdx+s,:] = complex(1.0, 1.0)
            else: # no SOC or SOC form QE.
                mask[:,idx:fdx,:]=complex(1.0, 1.0)

            #
            cs[:,:,:] = np.multiply(mask[:,:,:],v_k[k_index:k_index+1,:,:,0])

            total=0
            total+=np.sum(np.absolute(np.square((cs[:,:,bnd_idx]))))

            if(dim==3): # ploting for 3D system
                # we sum a very small part 0.0001 for ploting purpose.
                f.write (("%5.4f %5.4f %5.4f %5.4f \n") %(tau[n,0],tau[n,1],tau[n,2],total+0.0001))
            if(dim==2): # ploting for 2D system
                # we sum a very small part 0.0001 for ploting purpose.
                f.write (("%5.4f %5.4f %5.4f  \n") %(tau[n,0],tau[n,1],total+0.0001))

        f.close()

    #print("Fuck",site.shape)
    #print("Fuck",naw,e_k.shape)



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

def perturb_split(rot_op1,rot_op2,v_k,degen,return_v_k=False):
    import numpy as np
    from scipy import linalg as LAN

    op1 = np.dot(np.conj(v_k.T),np.dot(rot_op1,v_k))
    if len(degen)==0:
        op2 = np.dot(np.conj(v_k.T),np.dot(rot_op2,v_k))
        if return_v_k:
          return op1,op2,np.array([[]])
        else:
          return op1,op2

    v_k_temp = np.copy(v_k)
    
    for i in range(len(degen)):
        # degenerate subspace indices upper and lower lim
        ll = degen[i][0]
        ul = degen[i][-1]+1

        # diagonalize in degenerate subspace
        vals,weight = LAN.eigh(op1[ll:ul,ll:ul])

        # linear combination of eigenvectors of H that diagonalize
        v_k_temp[:,ll:ul]=np.dot(v_k_temp[:,ll:ul],weight)


    # return new operator in non degenerate basis
    op1 = np.dot(np.conj(v_k_temp.T),np.dot(rot_op1,v_k_temp))
    op2 = np.dot(np.conj(v_k_temp.T),np.dot(rot_op2,v_k_temp))

    if return_v_k:
      return(op1, op2, v_k_temp)
    else:
      return(op1, op2)



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_pdos ( data_controller, emin, emax, ne, delta ):

  arrays,attributes = data_controller.data_dicts()

  nawf = attributes['nawf']
  nspin = attributes['nspin']
  nktot = attributes['nkpnts']

  # PDOS calculation with gaussian smearing

  emax = np.amin(np.array([attributes['shift'], emax]))
  ene = np.linspace(emin, emax, ne)

  for ispin in range(nspin):

    pdosaux = np.zeros((nawf,ne), dtype=float)
    v_kaux = np.real(np.abs(arrays['v_k'][:,:,:,ispin])**2)

    E_k = arrays['E_k'][:,:,ispin]

    for n in range(ne):
      taux = np.exp(-((ene[n]-E_k)/delta)**2)/np.sqrt(np.pi)
      for m in range(nawf):
        pdosaux[m,n] += np.sum(taux*v_kaux[:,m,:])

    pdos = (np.zeros((nawf,ne),dtype=float) if rank==0 else None)

    comm.Reduce(pdosaux, pdos, op=MPI.SUM)
    pdosaux = None

    if rank == 0:
      pdos /= (float(nktot)*np.sqrt(np.pi)*delta)

    pdos_sum = (np.zeros(ne, dtype=float) if rank==0 else None)

    for m in range(nawf):
      if rank == 0:
        pdos_sum += pdos[m]
      fpdos = '%d_pdos_%d.dat'%(m,ispin)
      data_controller.write_file_row_col(fpdos, ene, (pdos[m] if rank==0 else None))

    fpdos = 'pdos_sum_%d.dat'%ispin
    data_controller.write_file_row_col(fpdos, ene, pdos_sum)


def do_pdos_adaptive ( data_controller, emin, emax, ne ):
  from .smearing import metpax, gaussian

  arrays = data_controller.data_arrays
  attributes = data_controller.data_attributes

  # PDoS Calculation with Gaussian Smearing
  emax = np.amin(np.array([attributes['shift'], emax]))
  ene = np.linspace(emin, emax, ne)

  nawf = attributes['nawf']

  for ispin in range(attributes['nspin']):

    E_k = np.real(arrays['E_k'][:,:,ispin])

    pdosaux = np.zeros((nawf,ne), dtype=float)

    v_kaux = np.real(np.abs(arrays['v_k'][:,:,:,ispin])**2)

    taux = np.zeros((arrays['deltakp'].shape[0],nawf), dtype=float)

    for n in range (ne):
      if attributes['smearing'] == 'gauss':
        taux = gaussian(ene[n], E_k, arrays['deltakp'][:,:,ispin]) 
      elif attributes['smearing'] == 'm-p':
        taux = metpax(ene[n], E_k, arrays['deltakp'][:,:,ispin])
      for i in range(nawf):
          # Adaptive Gaussian Smearing
          pdosaux[i,n] += np.sum(taux*v_kaux[:,i,:])

    pdos = (np.zeros((nawf,ne), dtype=float) if rank==0 else None)

    comm.Reduce(pdosaux, pdos, op=MPI.SUM)
    pdosaux = None

    if rank == 0:
      pdos /= float(attributes['nkpnts'])

    pdos_sum = (np.zeros(ne, dtype=float) if rank==0 else None)

    for m in range(nawf):
      if rank == 0:
        pdos_sum += pdos[m]
      fpdos = '%d_pdosdk_%d.dat'%(m,ispin)
      data_controller.write_file_row_col(fpdos, ene, (pdos[m] if rank==0 else None))

    fpdos = 'pdosdk_sum_%d.dat'%ispin
    data_controller.write_file_row_col(fpdos, ene, pdos_sum)



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np

def do_ortho ( Hks, Sks ):
  from scipy import linalg as spl
  from numpy import linalg as npl

  # If orthogonality is required, we have to apply a basis change to Hks as
  # Hks -> Sks^(-1/2)*Hks*Sks^(-1/2)

  nawf,_,nkpnts,nspin = Hks.shape
  S2k  = np.zeros((nawf,nawf,nkpnts), dtype=complex)
  for ik in range(nkpnts):
      S2k[:,:,ik] = npl.inv(spl.sqrtm(Sks[:,:,ik]))

  Hks_o = np.zeros((nawf,nawf,nkpnts,nspin), dtype=complex)
  for ispin in range(nspin):
      for ik in range(nkpnts):
          Hks_o[:,:,ik,ispin] = np.dot(S2k[:,:,ik], Hks[:,:,ik,ispin]).dot(S2k[:,:,ik])

  return Hks_o

def do_orthogonalize ( data_controller ):
  from scipy import fftpack as FFT

  arrays,attributes = data_controller.data_dicts()

  nktot = attributes['nkpnts']
  nawf,_,nk1,nk2,nk3,nspin = arrays['HRs'].shape

  if attributes['use_cuda']:
    from .cuda_fft import cuda_fftn
    arrays['Hks'] = cuda_fftn(np.moveaxis(arrays['HRs'],[0,1],[3,4]), axes=[0,1,2])
    arrays['Sks'] = cuda_fftn(np.moveaxis(arrays['SRs'],[0,1],[3,4]), axes=[0,1,2])
    arrays['Hks'] = np.reshape(np.moveaxis(arrays['Hks'],[3,4],[0,1]), (nawf,nawf,nktot,nspin), order='C')
    arrays['Sks'] = np.reshape(np.moveaxis(arrays['Sks'],[3,4],[0,1]), (nawf,nawf,nktot), order='C')
  else:
    arrays['Hks'] = FFT.fftn(arrays['HRs'],axes=[2,3,4])
    arrays['Sks'] = FFT.fftn(arrays['SRs'], axes=[2,3,4])
    arrays['Hks'] = np.reshape(arrays['Hks'], (nawf,nawf,nktot,nspin), order='C')
    arrays['Sks'] = np.reshape(arrays['Sks'], (nawf,nawf,nktot), order='C')

  arrays['Hks'] = do_ortho(arrays['Hks'], arrays['Sks'])
  arrays['Hks'] = np.reshape(arrays['Hks'], (nawf,nawf,nk1,nk2,nk3,nspin), order='C')
  arrays['Sks'] = np.reshape(arrays['Sks'], (nawf,nawf,nk1,nk2,nk3), order='C')
  if attributes['use_cuda']:
    from .cuda_fft import cuda_ifftn
    arrays['HRs'] = np.moveaxis(cuda_ifftn(np.moveaxis(arrays['Hks'],[0,1],[3,4]), axes=[0,1,2]),[3,4],[0,1])
  else:
    arrays['HRs'] = FFT.ifftn(arrays['Hks'], axes=[2,3,4])

  data_controller.broadcast_single_array('HRs')

  attributes['acbn0'] = False
  del arrays['SRs']



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_spin_Hall ( data_controller, twoD, do_ac ):
  from .perturb_split import perturb_split
  from .constants import ELECTRONVOLT_SI,ANGSTROM_AU,H_OVER_TPI,LL

  arry,attr = data_controller.data_dicts()

  s_tensor = arry['s_tensor']

  #-----------------------
  # Spin Hall calculation
  #-----------------------
  if attr['dftSO'] == False:
    if rank == 0:
      print('Relativistic calculation with SO required')
      comm.Abort()
    comm.Barrier()

  if rank == 0 and attr['verbose']:
    print('Writing bxsf files for Spin Berry Curvature')

  for n in range(s_tensor.shape[0]):
    ipol = s_tensor[n][0]
    jpol = s_tensor[n][1]
    spol = s_tensor[n][2]
    #----------------------------------------------
    # Compute the spin current operator j^l_n,m(k)
    #----------------------------------------------
    jdHksp = do_spin_current(data_controller, spol, ipol)

    jksp_is = np.empty_like(jdHksp)
    pksp_j = np.empty_like(jdHksp)

    for ik in range(jdHksp.shape[0]):
      for ispin in range(jdHksp.shape[3]):
        jksp_is[ik,:,:,ispin],pksp_j[ik,:,:,ispin] = perturb_split(jdHksp[ik,:,:,ispin], arry['dHksp'][ik,jpol,:,:,ispin], arry['v_k'][ik,:,:,ispin], arry['degen'][ispin][ik])
    jdHksp = None

    #---------------------------------
    # Compute spin Berry curvature... 
    #---------------------------------
    ene,shc,Om_k = do_Berry_curvature(data_controller, jksp_is, pksp_j)

    if rank == 0:
      if twoD:
        av0,av1 = arry['a_vectors'][0,:],arry['a_vectors'][1,:]
        cgs_conv = 1./(np.linalg.norm(np.cross(av0,av1))*attr['alat']**2)
      else:
        cgs_conv = 1.0e8*ANGSTROM_AU*ELECTRONVOLT_SI**2/(H_OVER_TPI*attr['omega'])
      shc *= cgs_conv

    cart_indices = (str(LL[spol]),str(LL[ipol]),str(LL[jpol]))

    fBerry = 'Spin_Berry_%s_%s%s.bxsf'%cart_indices
    nk1,nk2,nk3 = attr['nk1'],attr['nk2'],attr['nk3']
    Om_kps = (np.empty((nk1,nk2,nk3,2), dtype=float) if rank==0 else None)
    if rank == 0:
      Om_kps[:,:,:,0] = Om_kps[:,:,:,1] = Om_k[:,:,:]
    data_controller.write_bxsf(fBerry, Om_kps, 2)

    Om_k = Om_kps = None

    fshc = 'shcEf_%s_%s%s.dat'%cart_indices
    data_controller.write_file_row_col(fshc, ene, shc)

    ene = shc = None

    if do_ac:

      jdHksp = do_spin_current(data_controller, spol, ipol)

      jksp_js = np.empty_like(jdHksp)
      pksp_i = np.empty_like(jdHksp)

      for ik in range(jdHksp.shape[0]):
        for ispin in range(jdHksp.shape[3]):
          jksp_js[ik,:,:,ispin],pksp_i[ik,:,:,ispin] = perturb_split(jdHksp[ik,:,:,ispin], arry['dHksp'][ik,jpol,:,:,ispin], arry['v_k'][ik,:,:,ispin], arry['degen'][ispin][ik])
      jdHksp = None

      ene,sigxy = do_ac_conductivity(data_controller, jksp_js, pksp_i, ipol, jpol)
      if rank == 0:
        sigxy *= cgs_conv

      sigxyi = np.imag(ene*sigxy/105.4571) if rank==0 else None
      sigxyr = np.real(sigxy) if rank==0 else None
      sigxy = None

      fsigI = 'SCDi_%s_%s%s.dat'%cart_indices
      data_controller.write_file_row_col(fsigI, ene, sigxyi)

      fsigR = 'SCDr_%s_%s%s.dat'%cart_indices
      data_controller.write_file_row_col(fsigR, ene, sigxyr)


def do_anomalous_Hall ( data_controller, do_ac ):
  from .perturb_split import perturb_split
  from .constants import ELECTRONVOLT_SI,ANGSTROM_AU,H_OVER_TPI,LL

  arry,attr = data_controller.data_dicts()

  a_tensor = arry['a_tensor']

  #----------------------------
  # Anomalous Hall calculation
  #----------------------------
  if attr['dftSO'] == False:
    if rank == 0:
      print('Relativistic calculation with SO required')
      comm.Abort()
    comm.Barrier()

  if rank == 0 and attr['verbose']:
    print('Writing bxsf files for Berry Curvature')

  for n in range(a_tensor.shape[0]):
    ipol = a_tensor[n][0]
    jpol = a_tensor[n][1]

    dks = arry['dHksp'].shape

    pksp_i = np.zeros((dks[0],dks[2],dks[3],dks[4]),order="C",dtype=complex)
    pksp_j = np.zeros_like(pksp_i)

    for ik in range(dks[0]):
      for ispin in range(dks[4]):
        pksp_i[ik,:,:,ispin],pksp_j[ik,:,:,ispin] = perturb_split(arry['dHksp'][ik,ipol,:,:,ispin], arry['dHksp'][ik,jpol,:,:,ispin], arry['v_k'][ik,:,:,ispin], arry['degen'][ispin][ik])

    ene,ahc,Om_k = do_Berry_curvature(data_controller, pksp_i, pksp_j)

    if rank == 0:
      cgs_conv = 1.0e8*ANGSTROM_AU*ELECTRONVOLT_SI**2/(H_OVER_TPI*attr['omega'])

    cart_indices = (str(LL[ipol]),str(LL[jpol]))

    fBerry = 'Berry_%s%s.bxsf'%cart_indices
    nk1,nk2,nk3 = attr['nk1'],attr['nk2'],attr['nk3']
    Om_kps = (np.empty((nk1,nk2,nk3,2), dtype=float) if rank==0 else None)
    if rank == 0:
      Om_kps[:,:,:,0] = Om_kps[:,:,:,1] = Om_k[:,:,:]
    data_controller.write_bxsf(fBerry, Om_kps, 2)

    Om_k = Om_kps = None

    if rank == 0:
      ahc *= cgs_conv
    fahc = 'ahcEf_%s%s.dat'%cart_indices
    data_controller.write_file_row_col(fahc, ene, ahc)

    ene = ahc = None

    if do_ac:
      ene,sigxy = do_ac_conductivity(data_controller, pksp_i, pksp_j, ipol, jpol)
      if rank == 0:
        sigxy *= cgs_conv

      sigxyi = np.imag(ene*sigxy/105.4571) if rank==0 else None
      sigxyr = np.real(sigxy) if rank==0 else None
      sigxy = None

      fsigI = 'MCDi_%s%s.dat'%cart_indices
      data_controller.write_file_row_col(fsigI, ene, sigxyi)

      fsigR = 'MCDr_%s%s.dat'%cart_indices
      data_controller.write_file_row_col(fsigR, ene, sigxyr)

def do_Berry_curvature ( data_controller, jksp, pksp ):
  #----------------------
  # Compute spin Berry curvature
  #----------------------
  from .communication import gather_full
  from .smearing import intgaussian, intmetpax

  arrays,attributes = data_controller.data_dicts()

  snktot,nawf,_,nspin = pksp.shape
  fermi_up,fermi_dw = attributes['fermi_up'],attributes['fermi_dw']
  nk1,nk2,nk3 = attributes['nk1'],attributes['nk2'],attributes['nk3']

  # Compute only Omega_z(k)
  Om_znkaux = np.zeros((snktot,nawf), dtype=float)

  deltap = 0.05
  for ik in range(snktot):
    E_nm = (arrays['E_k'][ik,:,0] - arrays['E_k'][ik,:,0][:,None])**2 + deltap**2
    E_nm[np.where(E_nm<1.e-4)] = np.inf
    Om_znkaux[ik] = -2.0*np.sum(np.imag(jksp[ik,:,:,0]*pksp[ik,:,:,0].T)/E_nm, axis=1)
  E_nm = None

  attributes['emaxH'] = np.amin(np.array([attributes['shift'],attributes['emaxH']]))
  ### Hardcoded 'de'
  esize = 500
  ene = np.linspace(attributes['eminH'], attributes['emaxH'], esize)

  Om_zkaux = np.zeros((snktot,esize), dtype=float)

  for i in range(esize):
    if attributes['smearing'] == 'gauss':
      Om_zkaux[:,i] = np.sum((Om_znkaux[:,:]*intgaussian(arrays['E_k'][:,:,0],ene[i],arrays['deltakp'][:,:,0])), axis=1)
    elif attributes['smearing'] == 'm-p':
      Om_zkaux[:,i] = np.sum(Om_znkaux[:,:]*intmetpax(arrays['E_k'][:,:,0],ene[i],arrays['deltakp'][:,:,0]), axis=1)
    else:
      Om_zkaux[:,i] = np.sum(Om_znkaux[:,:]*(0.5 * (-np.sign(arrays['E_k'][:,:,0]-ene[i]) + 1)), axis=1)

  Om_zk = gather_full(Om_zkaux, attributes['npool'])
  Om_zkaux = None

  shc = None
  if rank == 0:
    shc = np.sum(Om_zk, axis=0)/float(attributes['nkpnts'])

  n0 = 0
  n = esize-1
  if rank == 0:
    for i in range(esize-1):
      if ene[i] <= fermi_dw and ene[i+1] >= fermi_dw:
        n0 = i
      if ene[i] <= fermi_up and ene[i+1] >= fermi_up:
        n = i
    Om_zk = np.reshape(Om_zk, (nk1,nk2,nk3,esize), order='C')
    Om_zk = Om_zk[:,:,:,n]-Om_zk[:,:,:,n0]

  return(ene, shc, Om_zk)

def do_ac_conductivity ( data_controller, jksp, pksp, ipol, jpol ):
  from .communication import gather_full
  from .smearing import intgaussian, intmetpax

  arry,attr = data_controller.data_dicts()

  # Compute the optical conductivity tensor sigma_xy(ene)

  ispin = 0

  emin = 0.0
  emax = attr['shift']
  ### Hardcode 'de'
  esize = 501
  ene = np.linspace(emin, emax, esize)

  sigxy_aux = smear_sigma_loop(data_controller, ene, jksp, pksp, ispin, ipol, jpol)

  sigxy = (np.zeros((esize),dtype=complex) if rank==0 else None)
  sigxyR = (np.zeros((esize),dtype=float) if rank==0 else None)
  sigxyI = (np.zeros((esize),dtype=float) if rank==0 else None)
  
  sigxy_auxR = np.ascontiguousarray(np.real(sigxy_aux))
  sigxy_auxI = np.ascontiguousarray(np.imag(sigxy_aux))
  
  comm.Reduce(sigxy_auxR, sigxyR, op=MPI.SUM)
  comm.Reduce(sigxy_auxI, sigxyI, op=MPI.SUM)

  sigxy_aux = sigxy_auxR = sigxy_auxI = None

  if rank == 0:
    sigxy = (sigxyR+1j*sigxyI)/float(attr['nkpnts'])
    return(ene, sigxy)
  else:
    return(None, None)

def smear_sigma_loop ( data_controller, ene, pksp_i, pksp_j, ispin, ipol, jpol ):
  from .smearing import intgaussian,intmetpax

  arry,attr = data_controller.data_dicts()

  esize = ene.size
  sigxy = np.zeros((esize), dtype=complex)

  snktot,nawf,_,nspin = pksp_j.shape
  f_nm = np.zeros((snktot,nawf,nawf), dtype=float)
  E_diff_nm = np.zeros((snktot,nawf,nawf), dtype=float)

  Ef = 0.0
  eps = 1.0e-16
  delta = 0.05

  if attr['smearing'] == None:
    fn = 1.0/(np.exp(arry['E_k'][:,:,ispin]/attr['temp'])+1)
  elif attr['smearing'] == 'gauss':
    fn = intgaussian(arry['E_k'][:,:,ispin], Ef, arry['deltakp'][:,:,ispin])
  elif smearing == 'm-p':
    fn = intmetpax(arry['E_k'][:,:,ispin], Ef, arry['deltakp'][:,:,ispin]) 

  # Collapsing the sum over k points
  for n in range(nawf):
    for m in range(nawf):
      if m != n:
        E_diff_nm[:,n,m] = (arry['E_k'][:,n,ispin]-arry['E_k'][:,m,ispin])**2
        f_nm[:,n,m] = (fn[:,n] - fn[:,m])*np.imag(pksp_j[:,n,m,ispin]*pksp_i[:,m,n,ispin])

  fn = None

  for e in range(esize):
    if attr['smearing'] != None:
      sigxy[e] = np.sum(f_nm[:,:,:]/(E_diff_nm[:,:,:]-(ene[e]+1.j*arry['deltakp2'][:,:nawf,:nawf,ispin])**2+eps))
    else:
      sigxy[e] = np.sum(f_nm[:,:,:]/(E_diff_nm[:,:,:]-(ene[e]+1.j*arry['delta'])**2+eps))

  F_nm = None
  E_diff_nm = None

  return np.nan_to_num(sigxy)

def do_spin_current ( data_controller, spol, ipol ):

  arry,attr = data_controller.data_dicts()

  Sj = arry['Sj'][spol]
  bnd = attr['bnd']
  snktot,_,nawf,nawf,nspin = arry['dHksp'].shape

  jdHksp = np.empty((snktot,nawf,nawf,nspin), dtype=complex)

  for ispin in range(nspin):
    for ik in range(snktot):
      jdHksp[ik,:,:,ispin] = 0.5*(np.dot(Sj,arry['dHksp'][ik,ipol,:,:,ispin])+np.dot(arry['dHksp'][ik,ipol,:,:,ispin],Sj))

  return jdHksp




#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


class ErrorHandler:

  def __init__ ( self ):
    pass

  def report_exception ( self, mname='UNKNOWN' ):
    '''
    Print the exception and traceback. Print suggestion if the error is recognized as a user mistake.

    Arguments:
        mname (str): Module name (name of function in PAOFLOW)

    Returns:
        None
    '''
    import sys
    import traceback
    #from .defs.module_prerequisites import key_error_strings
    from .defs.module_prerequisites import report_pre_reqs,module_pre_reqs

    etype, evalue, etb = sys.exc_info()
    print('Exception: ', etype)
    print(evalue)
    traceback.print_tb(etb)

    if etype is KeyError:
      pre_reqs = module_pre_reqs[mname] if mname in module_pre_reqs else '<%s>'%mname

      if len(pre_reqs) > 1:
        pr_str = ', '.join(pre_reqs[:-1]) + ' and %s'%pre_reqs[-1]
      else:
        pr_str = pre_reqs[0]

      print('')
      print(report_pre_reqs%(pr_str,mname))



# 
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

def do_adaptive_smearing ( data_controller, smearing ):
  from numpy.linalg import norm
  import numpy as np

  arrays,attributes = data_controller.data_dicts()

  #----------------------
  # adaptive smearing as in Yates et al. Phys. Rev. B 75, 195121 (2007).
  #----------------------

  a_vectors = arrays['a_vectors']

  nawf = attributes['nawf']
  nspin = attributes['nspin']
  nkpnts = attributes['nkpnts']
  npks = arrays['pksp'].shape[0]

  diag = np.diag_indices(nawf)

  dk = (8.*np.pi**3/attributes['omega']/(nkpnts))**(1./3.)

  afac = (1. if smearing=='m-p' else .7)

## DEV: Try to make contiguinuity conditional. Requires benchmark testing
  pksaux = np.ascontiguousarray(arrays['pksp'][:,:,diag[0],diag[1]])

  deltakp = np.zeros((npks,nawf,nspin), dtype=float)
  deltakp2 = np.zeros((npks,nawf,nawf,nspin), dtype=float)
  for n in range(nawf):
    deltakp[:,n] = norm(np.real(pksaux[:,:,n]), axis=1)
    for m in range(nawf):
      deltakp2[:,n,m,:] = norm(pksaux[:,:,n,:]-pksaux[:,:,m,:], axis=1)

  pksaux = None
  deltakp *= afac*dk
  deltakp2 *= afac*dk

  arrays['deltakp'] = deltakp
  arrays['deltakp2'] = deltakp2



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

def read_sh_nl (data_controller):
    # reads in shelks from pseudo files
    from os.path import join,exists
    import numpy as np


#    arry,attr = data_controller.data_dicts()
    arry = data_controller.data_arrays
    attr = data_controller.data_attributes
 
    # Get Shells for each species
    sdict = {}
    jchid = {}
    jchia = None

    for s in arry['species']:
      sdict[s[0]],jchid[s[0]] = read_pseudopotential(join(attr['workpath'],attr['savedir'],s[1]))


    for s,p in sdict.items():
        tmp_list=[]
        tmp_list_chi=[]
        for o in range(len(p)):
            tmp_list_chi.append(jchid[s][o])
            tmp_list.append(p[o])
            # if l=0 include it twice
            if p[o]==0 or len(jchid[s])==0:
                tmp_list_chi.append(jchid[s][o])
                tmp_list.append(p[o])

            # when using scalar rel pseido with spin orb..
            if len(jchid[s])==0:
                tmp=[]
                for o in sdict[s][::2]:
                    if o==0:
                        tmp.extend([0.5])
                    if o==1:
                        tmp.extend([0.5,1.5])
                    if o==2:
                        tmp.extend([1.5,2.5])
                    if o==3:
                        tmp.extend([2.5,3.5])
                jchid[s]=np.array(tmp)



        sdict[s] = np.array(tmp_list)
        jchid[s] = np.array(tmp_list_chi)



    # value of l
    shell   = np.hstack([sdict[a] for a in arry['atoms']])
    jchia   = np.hstack([jchid[a] for a in arry['atoms']])

    return shell[::2],jchia[::2]

############################################################################################
############################################################################################
############################################################################################

def read_pseudopotential ( fpp ):
  '''
  Reads a psuedopotential file to determine the included shells and occupations.

  Arguments:
      fnscf (string) - Filename of the pseudopotential, copied to the .save directory

  Returns:
      sh, nl (lists) - sh is a list of orbitals (s-0, p-1, d-2, etc)
                       nl is a list of occupations at each site
      sh and nl are representative of one atom only
  '''

  import numpy as np
  import xml.etree.cElementTree as ET
  import re
  from tempfile import NamedTemporaryFile

  sh = []
  jchi=[]

  # clean xnl before reading
  with open(fpp) as ifo:
      temp_str=ifo.read()

  temp_str = re.sub('&',' ',temp_str)
  f = NamedTemporaryFile(mode='w',delete=True)
  f.write(temp_str)

  try:
      iterator_obj = ET.iterparse(f.name,events=('start','end'))
      iterator     = iter(iterator_obj)
      event,root   = next(iterator)

      for event,elem in iterator:        
          try:
              for i in elem.findall("PP_PSWFC/"):
                  sh.append(int(i.attrib['l']))
          except Exception as e:
              pass

      sh   = np.array(sh)

  except Exception as e:
      with open(fpp) as ifo:
          ifs=ifo.read()
      res=re.findall("(.*)\s*Wavefunction",ifs)[1:]      
      sh=np.array(list(map(int,list([x.split()[1] for x in res]))))

  for i in elem.findall("PP_SPIN_ORB/"):
      try:
          jchi.append(float(i.attrib["jchi"]))
      except: pass

  jchi = np.array(jchi)

  return sh,jchi




# 
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
import time
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

def load_balancing ( size, rank, n ):
    # Load balancing
    splitsize = float(n)/float(size)
    start = int(np.around(rank*splitsize,decimals=2))
    stop = int(np.around((rank+1)*splitsize,decimals=2))
    return(start, stop)

# For each processor calculate 3 values:
# 0 - Total number of items to be scattered/gathered on this processor
# 1 - Index in complete array where the subarray begins
# 2 - Dimension of the subarray on this processor
def load_sizes ( size, n, dim):
    sizes = np.empty((size,3), dtype=int)
    splitsize = float(n)/float(size)
    for i in range(size):
        start = int(np.around(i*splitsize,decimals=2))
        stop = int(np.around((i+1)*splitsize,decimals=2))
        sizes[i][0] = dim*(stop-start)
        sizes[i][1] = dim*start
        sizes[i][2] = stop-start
    return sizes



# Scatters first dimension of an array of arbitrary length
def scatter_array ( arr, sroot=0 ):

    # Compute data type and shape of the scattered array on this process
    pydtype = None
    auxlen = None

    # An array to store the size and dimensions of scattered arrays
    lsizes = np.empty((size,3), dtype=int)
    if rank == sroot:
        pydtype = arr.dtype
        auxshape = np.array(list(arr.shape))
        auxlen = len(auxshape)
        lsizes = load_sizes(size, arr.shape[0], np.prod(arr.shape[1:]))

    # Broadcast the data type and dimension of the scattered array
    pydtype = comm.bcast(pydtype, root=sroot)
    auxlen = comm.bcast(auxlen, root=sroot)

    # An array to store the shape of array's dimensions
    if rank != sroot:
        auxshape = np.zeros((auxlen,), dtype=int)

    # Broadcast the shape of each dimension
    for i in np.arange(auxlen):
        auxshape[i] = comm.bcast(auxshape[i], root=sroot)

    comm.Bcast([auxshape, MPI.INT], root=sroot)
    comm.Bcast([lsizes, MPI.INT], root=sroot)

    # Change the first dimension of auxshape to the correct size for scatter
    auxshape[0] = lsizes[rank][2]

    # Initialize aux array
    arraux = np.empty(auxshape, dtype=pydtype)

    # Get the datatype for the MPI transfer
    mpidtype = MPI._typedict[np.dtype(pydtype).char]

    # Scatter the data according to load_sizes
    comm.Scatterv([arr, lsizes[:,0], lsizes[:,1], mpidtype], [arraux, mpidtype], root=sroot)

    return arraux

# Gathers first dimension of an array of arbitrary length
def gather_array ( arr, arraux, sroot=0 ):

    # An array to store the size and dimensions of gathered arrays
    lsizes = np.empty((size,3), dtype=int)
    if rank == sroot:
        lsizes = load_sizes(size, arr.shape[0], np.prod(arr.shape[1:]))

    # Broadcast the data offsets
    comm.Bcast([lsizes, MPI.INT], root=sroot)

    # Get the datatype for the MPI transfer
    mpidtype = MPI._typedict[np.dtype(arraux.dtype).char]

    # Gather the data according to load_sizes
    comm.Gatherv([arraux, mpidtype], [arr, lsizes[:,0], lsizes[:,1], mpidtype], root=sroot)


def scatter_full(arr,npool,sroot=0):

    if rank==sroot:
        nsizes = comm.bcast(arr.shape)
    else:
        nsizes = comm.bcast(None)

    comm.Barrier()

    nsize=nsizes[0]
    full = int(nsize/size)

    ts,te = load_balancing(size,rank,nsize%size)
    full+=te-ts

    if len(nsizes)>1:
        per_proc_shape = np.concatenate((np.array([full]),
                                         nsizes[1:]))
    else: per_proc_shape = np.array([full])

    pydtype=None
    if rank==sroot:
        pydtype=arr.dtype
    pydtype = comm.bcast(pydtype)

    temp = np.zeros(per_proc_shape,order="C",dtype=pydtype)

    nchunks = nsize//size
    
    if nchunks!=0:
        for pool in range(npool):
            chunk_s,chunk_e = load_balancing(npool,pool,nchunks)

            if rank==sroot:
                temp[chunk_s:chunk_e] = scatter_array(np.ascontiguousarray(arr[(chunk_s*size):(chunk_e*size)]))
            else:
                temp[chunk_s:chunk_e] = scatter_array(None)
    else:
        chunk_e=0

    if nsize%size!=0:
        if rank==sroot:
            temp[chunk_e:] = scatter_array(np.ascontiguousarray(arr[(chunk_e*size):]))
        else:
            temp[chunk_e:] = scatter_array(None)

    return temp


def gather_full(arr,npool,sroot=0):

    first_ind_per_proc = np.array([arr.shape[0]])
    nsize              = np.zeros_like(first_ind_per_proc)

    comm.Barrier()
    comm.Allreduce(first_ind_per_proc,nsize)

    if len(arr.shape)>1:
        per_proc_shape = np.concatenate((nsize,arr.shape[1:]))
    else: per_proc_shape = np.array([arr.shape[0]])

    nsize=nsize[0]

    if rank==sroot:
        temp = np.zeros(per_proc_shape,order="C",dtype=arr.dtype)
    else: temp = None

    nchunks = nsize//size
    
    if nchunks!=0:
        for pool in range(npool):
            chunk_s,chunk_e = load_balancing(npool,pool,nchunks)

            if rank==sroot:
                gather_array(temp[(chunk_s*size):(chunk_e*size)],arr[chunk_s:chunk_e],sroot=sroot)
            else:
                gather_array(None,arr[chunk_s:chunk_e],sroot=sroot)
    else:
        chunk_e=0

    if nsize%size!=0:
        if rank==sroot:
            gather_array(temp[(chunk_e*size):],arr[chunk_e:],sroot=sroot)
        else:
            gather_array(None,arr[chunk_e:],sroot=sroot)
        
    if rank == sroot:
        return temp


def gather_scatter(arr,scatter_axis,npool):
    #scatter indices for scatter_axis to each proc
    axis_ind = np.array(list(range(arr.shape[scatter_axis])),dtype=int)
    axis_ind = scatter_full(axis_ind,npool)

    #broadcast indices that for scattered array to proc with rank 'r'
    size_r = np.zeros((size),dtype=int,order='C')
    scatter_ind = np.zeros((arr.shape[scatter_axis]),dtype=int,order='C')
    
    if rank==0:
        gather_array(size_r,np.array(axis_ind.size,dtype=int))
        gather_array(scatter_ind,np.array(axis_ind,dtype=int))
    else:
        gather_array(None,np.array(axis_ind.size,dtype=int))
        gather_array(None,np.array(axis_ind,dtype=int))

    axis_ind = None

    comm.Bcast(size_r)
    comm.Bcast(scatter_ind)

    #start and end points of indices of scatter axis for each proc
    end   = np.cumsum(size_r)
    start = end - size_r
    size_r = None
    
    for r in range(size):
        comm.Barrier()
        #gather array from each proc with indices for each proc on scatter_axis
        if r == rank:
            temp = gather_full(np.take(arr,scatter_ind[start[r]:end[r]],axis=scatter_axis),npool,sroot=r)
        else:
            gather_full(np.take(arr,scatter_ind[start[r]:end[r]],axis=scatter_axis),npool,sroot=r)

    start = end = scatter_ind = None

    return temp


def gen_window(array,root=0):
    # creates a shared memory copy of array on
    # rank == root that all procs can access

    if rank==root:
        array_shape=array.shape
        pydtype = array.dtype
    else:
        array_shape=None
        pydtype=None

    array_shape=comm.bcast(array_shape)
    pydtype=comm.bcast(pydtype)

    size=np.prod(array_shape)

    itemsize = MPI._typedict[np.dtype(pydtype).char].Get_size()
    if rank == root:
        nbytes = size * itemsize
    else:
        nbytes = 0

    win = MPI.Win.Allocate_shared(nbytes, itemsize, comm=comm)
    buf, itemsize = win.Shared_query(0)
    win_array = np.ndarray(buffer=buf, dtype=pydtype, shape=array_shape,)

    if rank==root:
        win_array[:]=array

    comm.Barrier()
    
    return win_array



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016,2017 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from constants import ANGSTROM_AU

def doubling_HRs ( data_controller ):
    from scipy.fftpack import fftshift
    from mpi4py import MPI

    if MPI.COMM_WORLD.Get_rank() != 0:
       return

    arry,attr = data_controller.data_dicts()

    do_spin_orbit = attr['do_spin_orbit']    
    nx = attr['nx']   
    ny = attr['ny']   
    nz = attr['nz']   
    nk1 = attr['nk1']   
    nk2 = attr['nk2']   
    nk3 = attr['nk3']   
    nspin = attr['nspin']   

    nkpts=nk1*nk2*nk3

    cell_index=np.zeros((nk1,nk2,nk3,3),dtype=int)
    new_index=np.zeros((3,nkpts),dtype=int)
    
    for i in range(nk1):
        for j in range(nk2):
            for k in range(nk3):
                n = k + j*nk3 + i*nk2*nk3
                Rx = float(i)/float(nk1)
                Ry = float(j)/float(nk2)
                Rz = float(k)/float(nk3)
                if Rx >= 0.5: Rx=Rx-1.0
                if Ry >= 0.5: Ry=Ry-1.0
                if Rz >= 0.5: Rz=Rz-1.0
                Rx -= int(Rx)
                Ry -= int(Ry)
                Rz -= int(Rz)
                # the minus sign in Rx*nk1 is due to the Fourier transformation (Ri-Rj)
                ix=int(-round(Rx*nk1,0))
                iy=int(-round(Ry*nk2,0))
                iz=int(-round(Rz*nk3,0))
                
                cell_index[ix,iy,iz,0]=i
                cell_index[ix,iy,iz,1]=j
                cell_index[ix,iy,iz,2]=k           

                new_index[0,n]=ix
                new_index[1,n]=iy
                new_index[2,n]=iz

    # This construction is doubling along the X direction nx times    
    for dx in range(nx):
        
        HR_double= np.zeros((2*attr['nawf'],2*attr['nawf'],nk1,nk2,nk3,nspin),dtype=complex)
        for ix in range(min(new_index[0,:]),max(new_index[0,:])+1):
            for iy in range(min(new_index[1,:]),max(new_index[1,:])+1):
                for iz in range(min(new_index[2,:]),max(new_index[2,:])+1):
                                  
                    i,j,k = cell_index[ix,iy,iz,:] # doubled cell index

                    if ( ((2*ix) >= min(new_index[0,:])) and ((2*ix) <= (max(new_index[0,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[2*ix,iy,iz,:]                
                        # Upper left HR_double block                             
                        HR_double[0:attr['nawf'],0:attr['nawf'],i,j,k,:]           = arry['HRs'][:,:,m,n,l,:]
                        # Lower right HR_double block                             
                        HR_double[attr['nawf']:2*attr['nawf'],attr['nawf']:2*attr['nawf'],i,j,k,:] = arry['HRs'][:,:,m,n,l,:]
                
                    if ( ((2*ix+1) >= min(new_index[0,:])) and ((2*ix+1) <= (max(new_index[0,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[2*ix+1,iy,iz,:]
                        #Upper right HR_double block                
                        HR_double[0:attr['nawf'],attr['nawf']:2*attr['nawf'],i,j,k,:] = arry['HRs'][:,:,m,n,l,:]
                
                    if ( ((2*ix-1) >= min(new_index[0,:])) and ((2*ix-1) <= (max(new_index[0,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[2*ix-1,iy,iz,:]
                        #Lower left HR_double block
                        HR_double[attr['nawf']:2*attr['nawf'],0:attr['nawf'],i,j,k,:] = arry['HRs'][:,:,m,n,l,:]

        arry['HRs'] = HR_double
        HR_double = None
        arry['tau']   = np.append(arry['tau'],arry['tau'][:,:]+arry['a_vectors'][0,:]*ANGSTROM_AU,axis=0)
        arry['a_vectors'][0,:]=2*arry['a_vectors'][0,:]
        doubling_attr_arry(data_controller)



    # This construction is doubling along the X direction nx times    
    for dy in range(ny):
        HR_double= np.zeros((2*attr['nawf'],2*attr['nawf'],nk1,nk2,nk3,nspin),dtype=complex)

        for ix in range(min(new_index[0,:]),max(new_index[0,:])+1):
            for iy in range(min(new_index[1,:]),max(new_index[1,:])+1):
                for iz in range(min(new_index[2,:]),max(new_index[2,:])+1):
                                  
                    i,j,k = cell_index[ix,iy,iz,:] # doubled cell index

                    if ( ((2*iy) >= min(new_index[1,:])) and ((2*iy) <= (max(new_index[1,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[ix,2*iy,iz,:]                
                        # Upper left HR_double block                             
                        HR_double[0:attr['nawf'],0:attr['nawf'],i,j,k,:]           =  arry['HRs'][:,:,m,n,l,:]
                        # Lower right HR_double block                             
                        HR_double[attr['nawf']:2*attr['nawf'],attr['nawf']:2*attr['nawf'],i,j,k,:] =  arry['HRs'][:,:,m,n,l,:]
                
                    if ( ((2*iy+1) >= min(new_index[1,:])) and ((2*iy+1) <= (max(new_index[1,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[ix,2*iy+1,iz,:]
                        #Upper right HR_double block                
                        HR_double[0:attr['nawf'],attr['nawf']:2*attr['nawf'],i,j,k,:] =  arry['HRs'][:,:,m,n,l,:]
                
                    if ( ((2*iy-1) >= min(new_index[1,:])) and ((2*iy-1) <= (max(new_index[1,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[ix,2*iy-1,iz,:]
                        #Lower left HR_double block                
                        HR_double[attr['nawf']:2*attr['nawf'],0:attr['nawf'],i,j,k,:] =  arry['HRs'][:,:,m,n,l,:]
                

        arry['HRs'] = HR_double
        HR_double = None
        arry['tau']   = np.append(arry['tau'],arry['tau'][:,:]+arry['a_vectors'][1,:]*ANGSTROM_AU,axis=0)
        arry['a_vectors'][1,:]=2*arry['a_vectors'][1,:]
        doubling_attr_arry(data_controller)

    # This construction is doubling along the X direction nx times    
    delete_index=0
    for dz in range(nz):

        HR_double= np.zeros((2*attr['nawf'],2*attr['nawf'],nk1,nk2,nk3,nspin),dtype=complex)
        
        for ix in range(min(new_index[0,:]),max(new_index[0,:])+1):
            for iy in range(min(new_index[1,:]),max(new_index[1,:])+1):
                for iz in range(min(new_index[2,:]),max(new_index[2,:])+1):
                                  
                    i,j,k = cell_index[ix,iy,iz,:] # doubled cell index

                    if ( ((2*iz) >= min(new_index[2,:])) and ((2*iz) <= (max(new_index[2,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[ix,iy,2*iz,:]                
                        # Upper left HR_double block                             
                        HR_double[0:attr['nawf'],0:attr['nawf'],i,j,k,:]           = arry['HRs'][:,:,m,n,l,:]
                        # Lower right HR_double block                             
                        HR_double[attr['nawf']:2*attr['nawf'],attr['nawf']:2*attr['nawf'],i,j,k,:] = arry['HRs'][:,:,m,n,l,:]
                
                    if ( ((2*iz+1) >= min(new_index[2,:])) and ((2*iz+1) <= (max(new_index[2,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[ix,iy,2*iz+1,:]
                        #Upper right HR_double block                
                        HR_double[0:attr['nawf'],attr['nawf']:2*attr['nawf'],i,j,k,:] = arry['HRs'][:,:,m,n,l,:]
                
                    if ( ((2*iz-1) >= min(new_index[2,:])) and ((2*iz-1) <= (max(new_index[2,:])))):
                        i,j,k = cell_index[ix,iy,iz,:]
                        m,n,l = cell_index[ix,iy,2*iz-1,:]
                        #Lower left HR_double block                
                        HR_double[attr['nawf']:2*attr['nawf'],0:attr['nawf'],i,j,k,:] = arry['HRs'][:,:,m,n,l,:]
        
        arry['HRs'] = HR_double
        HR_double = None
        arry['tau']   = np.append(arry['tau'],arry['tau'][:,:]+arry['a_vectors'][2,:]*ANGSTROM_AU,axis=0)
        arry['a_vectors'][1,:]=2*arry['a_vectors'][1,:]
        doubling_attr_arry(data_controller)


def doubling_attr_arry ( data_controller ):

    arry,attr = data_controller.data_dicts()

    do_spin_orbit = attr['do_spin_orbit']    

    # Increassing nawf/natoms
    attr['nawf'] = 2*attr['nawf']
    attr['natoms'] = 2*attr['natoms']
    #doubling the atom number of orbitals / orbital character / multiplicity
    arry['naw']   = np.append(arry['naw'],arry['naw']) 
    arry['sh']   = np.append(arry['sh'],arry['sh']) 
    arry['nl']   = np.append(arry['nl'],arry['nl']) 
    # If the SOC is included pertubative
    if (attr['do_spin_orbit']):
        arry['lambda_p']   = np.append(arry['lambda_p'],arry['lambda_p'])
        arry['lambda_d']   = np.append(arry['lambda_d'],arry['lambda_d'])
        arry['orb_pseudo'] = np.append(arry['orb_pseudo'],arry['orb_pseudo'])
#






def do_band_curvature ( data_controller ):
    '''
    Calculate the Gradient of the k-space Hamiltonian, 'Hksp'
    Requires 'Hksp' and 'pksp'
    Yields 'd2Hksp'

    Arguments:
        None

    Returns:
        None
    '''

    from .do_d2Hd2k import do_d2Hd2k_ij
    import numpy as np
    from .perturb_split import perturb_split

    ary,attr = data_controller.data_dicts()
    bnd = attr['bnd']    
    nawf = attr['nawf']    
    E_k = ary['E_k']    

    # not really the inverse mass tensor..it's actually tksp
    # but we are calling it d2Ed2k for now to save memory.
    d2Ed2k,dvec_list = do_d2Hd2k_ij(ary['Hksp'],ary['Rfft'],attr['alat'],
                                    attr['npool'],ary['v_k'],
                                    bnd,ary['degen'])


    
    # d2Ed2k is only the 6 unique components of the curvature  
    # (inverse effective mass ) tensor. This is one to save memory.
    ij_ind = np.array([[0,0],[1,1],[2,2],[0,1],[1,2],[0,2]],dtype=int)
    E_temp = np.zeros((bnd,nawf),order="C")

    #----------------------
    # for d2E/d2k_ij
    #----------------------
    for ispin in range(d2Ed2k.shape[3]):
      for ik in range(d2Ed2k.shape[1]):          

        # tksp_ij = <psi|d2Hd2k_ij|psi>

        # ij component of second derivative of the energy is:
        # tksp_ij + sum_i( (pksp_i*pksp_j.T + pksp_j*pksp_i.T)/(E_i-E_j) )
        E_temp = ((E_k[ik,:,ispin]-E_k[ik,:,ispin][:,None])[:,:]).T
        E_temp[np.where(np.abs(E_temp)<1.e-5)]=np.inf

        for ij in range(ij_ind.shape[0]):
            ipol = ij_ind[ij,0]
            jpol = ij_ind[ij,1]

            # to avoid a zero in the denominator when E_i=E_j
            if dvec_list[ij][ispin][ik].size:
                v_k=dvec_list[ij][ispin][ik]
            else:
                v_k=ary['v_k'][ik,:,:,ispin]
            
            pksp_i=np.conj(v_k.T).dot(ary['dHksp'][ik,ipol,:,:,ispin]).dot(v_k)
            pksp_j=np.conj(v_k.T).dot(ary['dHksp'][ik,jpol,:,:,ispin]).dot(v_k)
            #pksp_i=ary['pksp'][ik,ipol,:,:,ispin]
            #pksp_j=ary['pksp'][ik,jpol,:,:,ispin]
            # this is where d2Ed2k becomes the actual curvature tensor
            d2Ed2k[ij,ik,:,ispin] += np.sum((((pksp_i*pksp_j.T +\
                                               pksp_j*pksp_i.T) / E_temp).real),axis=1)[:bnd]


    ary['d2Ed2k']=d2Ed2k
      
              



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
#  M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
#  PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
#  Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
from mpi4py import MPI

def build_Pn ( nawf, nbnds, nkpnts, nspin, U ):
  Pn = 0.
  for ispin in range(nspin):
    for ik in range(nkpnts):
      UU = np.transpose(U[:,:,ik,ispin]) #transpose of U. Now the columns of UU are the eigenvector of length nawf
      Pn += np.real(np.sum(np.conj(UU)*UU,axis=0))/nkpnts/nspin
  return Pn


def do_projectability ( data_controller ):

  #----------------------
  # Building the Projectability
  #----------------------
  rank = MPI.COMM_WORLD.Get_rank()

  arry,attr = data_controller.data_dicts()

  pthr,shift = attr['pthr'],attr['shift']

  if rank != 0:
    attr['shift'] = None
  else:
    Pn = build_Pn(attr['nawf'], attr['nbnds'], attr['nkpnts'], attr['nspin'], arry['U'])

    if attr['verbose']:
      print('Projectability vector ', Pn)

    # Check projectability and decide bnd
    bnd = 0
    for n in range(attr['nbnds']):
      if Pn[n] > attr['pthr']:
        bnd += 1
    Pn = None
    attr['bnd'] = maxbnd = bnd
    if bnd == attr['nawf']:
      maxbnd = bnd-1
      print('WARNING: All bands meet the projectability threshold. Consider increasing nbnd in QE.')

    if 'shift' not in attr or attr['shift']=='auto':
      shift_v = np.max([np.amin(arry['my_eigsmat'][maxbnd,:,:]), np.amax(arry['my_eigsmat'][bnd-1,:,:])])
      attr['shift'] = (shift_v if shift=='auto' else shift)

    if attr['verbose']:
      print('# of bands with good projectability > {} = {}'.format(attr['pthr'],bnd))
    if attr['verbose'] and bnd < attr['nbnds']:
      print('Range of suggested shift ', np.amax(arry['my_eigsmat'][bnd-1,:,:]), ' , ', np.amax(arry['my_eigsmat'][maxbnd,:,:]))

  # Broadcast 
  data_controller.broadcast_attribute('bnd')
  data_controller.broadcast_attribute('shift')



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


def do_non_ortho ( Hks, Sks ):
  import numpy as np
  from scipy import linalg as spl

  # Take care of non-orthogonality, if needed
  # Hks from projwfc is orthogonal. If non-orthogonality is required, we have to apply a basis change to Hks as
  # Hks -> Sks^(1/2)*Hks*Sks^(1/2)

  nawf,_,nkpnts,nspin = Hks.shape
  S2k  = np.zeros((nawf,nawf,nkpnts), dtype=complex)
  for ik in range(nkpnts):
    S2k[:,:,ik] = spl.sqrtm(Sks[:,:,ik])

  Hks_no = np.zeros((nawf,nawf,nkpnts,nspin), dtype=complex)
  for ispin in range(nspin):
    for ik in range(nkpnts):
      Hks_no[:,:,ik,ispin] = np.dot(S2k[:,:,ik], Hks[:,:,ik,ispin]).dot(S2k[:,:,ik])

  return Hks_no



# *************************************************************************************
# *                                                                                   *
# *   PAOFLOW *  Marco BUONGIORNO NARDELLI * University of North Texas 2016-2018      *
# *                                                                                   *
# *************************************************************************************
#
#  Copyright 2016-2018 - Marco BUONGIORNO NARDELLI (mbn@unt.edu) - AFLOW.ORG consortium
#
#  This file is part of AFLOW software.
#
#  AFLOW is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# *************************************************************************************

import numpy as np
from scipy.constants import hbar
from scipy.constants import Boltzmann as kb

me = 9.10938e-31
e = 1.60217662e-19
ev2j = 1.60217662e-19
epso = 8.854187817e-12

def acoustic_model ( temp, eigs, params ):
  # Formula from fiorentini paper on Mg3Sb2
  temp *= ev2j
  E = eigs * ev2j # Eigenvalues in J
  v = params['v'] # Velocity in m/s
  rho = params['rho'] # Mass density kg/m^3
  ms = params['ms']*me #effective mass tensor in kg 
  D_ac = params['D_ac']*ev2j # Acoustic deformation potential in J

  return (2*np.pi*rho*(hbar**2*v)**2)/(2*ms)**1.5*(D_ac**2)*np.sqrt(E)*temp


def optical_model ( temp, eigs, params ):
  # Formula from jacoboni theory of electron transport in semiconductors
  temp *= ev2j
  E = eigs * ev2j
  hwlo = np.array(params['hwlo'])*ev2j # Phonon freq
  rho = params['rho'] # Mass density kg/m^3
  D_op = params['D_op']*ev2j # Acoustic deformation potential in J
  ms = params['ms']*me #effective mass tensor in kg 

  x = E/temp
  x0 = hwlo/temp
  X = x - x0
  X[X<0] = 0

  Nop = 1 / (np.exp(x0)-1)

  return (np.sqrt(2*temp)*np.pi*x0*rho*hbar**2)/((ms**1.5)*(D_op**2)*(Nop*np.sqrt(x+x0)+(Nop+1)*np.sqrt(X)))

def polar_acoustic_model ( temp, eigs, params ):

  temp *= ev2j
  E = eigs * ev2j
  piezo = params['piezo']  # Piezoelectric constant
  nd = params['doping_conc']*1e6 # Doping concentration in /m^3
  eps_0 = params['eps_0']*epso # Low freq dielectric const
  eps_inf = params['eps_inf']*epso # High freq dielectirc const
  ms = params['ms']*me #effective mass tensor in kg 
  rho = params['rho']
  v = params['v']

  eps = eps_inf + eps_0
  qo = np.sqrt(abs(nd)*e**2/(eps*temp))
  eps_o = ((hbar*qo)**2)/(2*ms)
  P_pac = (((piezo*e)**2*ms**0.5*temp)/(np.sqrt(2*E)*2*np.pi*eps**2*hbar**2*rho*v**2))*(1-(eps_o/(2*E))*np.log(1+4*E/eps_o)+1/(1+4*E/eps_o))
  P_pac[np.isnan(P_pac)] = 0
  return 1 / P_pac

def polar_optical_model ( temp, eigs, params ):
  # Formula from fiorentini paper on Mg3Sb2
  temp *= ev2j
  E = eigs * ev2j
  Ef = params['Ef']*ev2j #fermi energy
  hwlo = np.array(params['hwlo'])*ev2j # Phonon freq
  eps_0 = params['eps_0']*epso #low freq dielectric const
  eps_inf = params['eps_inf']*epso #high freq dielectirc const
  ms = params['ms']*me #effective mass tensor in kg


  fermi = lambda E,Ef,T : 1/(np.exp((E-Ef)/T)+1.)
  planck = lambda hwlo,T : 1/(np.exp(hwlo/T)-1)

  P_pol = 0.
  eps = eps_inf + eps_0
  eps_inv = 1/eps_inf - 1/eps

  for hw in hwlo:
    f = fermi(E, Ef, temp)
    fp = fermi(E+hw, Ef, temp)
    fm = fermi(E-hw, Ef, temp)
    n = planck(hw, temp)
    Wo = e**2/(4*np.pi*hbar**2) * np.sqrt(2*ms*hw)*eps_inv
    Z = 2/(Wo*np.sqrt(hw))

    def remove_NaN ( arr ):
      arr[np.isnan(arr)] = 0
      return arr

    A = remove_NaN((n+1)*fp/f * ((2*E+hw)*np.arcsinh(np.sqrt(E/hw))-np.sqrt(E*(E+hw))))
    B = remove_NaN(np.heaviside(E-hw,1)*n*fm/f * ((2*E-hw)*np.arccosh(np.sqrt(E/hw))-np.sqrt(E*(E-hw))))
    t1 = remove_NaN((n+1)*fp/f * np.arcsinh(np.sqrt(E/hw)))
    t2 = remove_NaN(np.heaviside(E-hw,1)*n*fm/f * np.arccosh(np.sqrt(E/hw)))
    C = 2*E*(t1+t2)
    P = (C-A-B)/(Z*E**1.5)
    P_pol += P

  return 1 / P_pol


def impurity_model ( temp, eigs, params ):
  #formula from fiorentini paper on Mg3Sb2
  temp *= ev2j
  E = eigs * ev2j
  nI = params['nI']*1e6 # impurity conc in /m^3
  Zi = params['Zi']
  ms = params['ms']*me #effective mass tensor in kg 
  eps_0 = params['eps_0']*epso #low freq dielectric const
  eps_inf = params['eps_inf']*epso #high freq dielectirc const

  eps = eps_inf+eps_0
  qo = np.sqrt(e**2*nI/(eps*temp))
  x = (hbar*qo)**2/(8*ms*E)
  P_imp = np.pi*nI*Zi**2*e**4/(E**1.5*np.sqrt(2*ms)*(4*np.pi*eps)**2)
  return 1 / (P_imp * (np.log(1+1./x)-1./(1+x)))

def builtin_tau_model ( label, params, weight ):
  from .TauModel import TauModel

  model = TauModel(params=params, weight=weight)

  if label == 'acoustic':
    model.function = acoustic_model
  elif label == 'optical':
    model.function = optical_model
  elif label == 'polar_optical':
    model.function = polar_optical_model
  elif label == 'polar_acoustic':
    model.function = polar_acoustic_model
  elif label == 'impurity':
    model.function = impurity_model
  else:
    print('Model not implemented.')
    return None

  return model
